#  Copyright (c) 2021. Davi Pereira dos Santos
#  This file is part of the ldict project.
#  Please respect the license - more about this in the section (*) below.
#
#  ldict is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ldict is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with ldict.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is illegal and unethical regarding the effort and
#  time spent here.
from abc import abstractmethod, ABC
from collections import UserDict
from random import Random
from typing import Dict, TypeVar

from ldict.exception import WrongKeyType

VT = TypeVar("VT")


class AbstractLazyDict(UserDict, Dict[str, VT]):
    """
    >>> from ldict.core.ldict_ import Ldict
    >>> from ldict.frozenlazydict import FrozenLazyDict
    >>> {"x": 5} == Ldict({"x": 5})
    True
    >>> {"w": 5} == Ldict({"x": 5})
    False
    >>> {"x": 4} == Ldict({"x": 5})
    False
    >>> {"x": 5} == FrozenLazyDict({"x": 5})
    True
    >>> {"w": 5} == FrozenLazyDict({"x": 5})
    False
    >>> {"x": 4} == FrozenLazyDict({"x": 5})
    False
    """

    rnd: Random

    @property
    @abstractmethod
    def asdict(self):  # pragma: no cover
        raise NotImplementedError

    @property
    def evaluated(self) -> "AbstractLazyDict":
        self.evaluate()
        return self

    def evaluate(self):
        """
        >>> from ldict import ldict
        >>> f = lambda x: {"y": x+2}
        >>> d = ldict(x=3)
        >>> a = d >> f
        >>> a
        {
            "x": 3,
            "y": "→(x)"
        }
        >>> a.evaluate()
        >>> a
        {
            "x": 3,
            "y": 5
        }
        """
        for field in self:
            v = self[field]
            if isinstance(v, AbstractLazyDict):
                v.evaluate()

    def __ne__(self, other):
        return not (self == other)


class AbstractMutableLazyDict(AbstractLazyDict, ABC):
    frozen: AbstractLazyDict

    @property
    def rnd(self):
        return self.frozen.rnd

    @property
    def data(self):
        return self.frozen.data

    def __getitem__(self, item):
        return self.frozen[item]

    def __setitem__(self, key: str, value):
        if not isinstance(key, str):  # pragma: no cover
            raise WrongKeyType(f"Key must be string, not {type(key)}.", key)
        self.frozen >>= {key: value}

    def __getattr__(self, item):
        if item in self.frozen:
            return self.frozen[item]
        return self.__getattribute__(item)

    def __repr__(self):
        return repr(self.frozen)

    def __str__(self):
        return str(self.frozen)

    @property
    def asdict(self):
        """
        >>> from ldict import ldict
        >>> d = ldict(x=3, y=5)
        >>> ldict(x=7, y=8, d=d).asdict
        {'x': 7, 'y': 8, 'd': {'x': 3, 'y': 5}}
        """
        return self.frozen.asdict

    def __eq__(self, other):
        return self.frozen == other



#  Copyright (c) 2021. Davi Pereira dos Santos
#  This file is part of the ldict project.
#  Please respect the license - more about this in the section (*) below.
#
#  ldict is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ldict is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with ldict.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is illegal and unethical regarding the effort and
#  time spent here.
from json import JSONEncoder


class CustomJSONEncoder(JSONEncoder):
    """
    >>> from ldict.frozenlazydict import FrozenLazyDict
    >>> ldict = FrozenLazyDict
    >>> a = ldict(x=3)
    >>> ldict(d=a, y=5)
    {
        "d": {
            "x": 3
        },
        "y": 5
    }
    >>> from pandas.core.frame import DataFrame, Series
    >>> df = DataFrame([[1,2],[3,4]])
    >>> df
       0  1
    0  1  2
    1  3  4
    >>> b = ldict(d=a, y=5, df=df, ell=...)
    >>> b
    {
        "d": {
            "x": 3
        },
        "y": 5,
        "df": {0: {0: 1, 1: 3}, 1: {0: 2, 1: 4}},
        "ell": "..."
    }
    >>> from numpy import array
    >>> ldict(b=b, z=9, c=(c:=array([1,2,3])), d=Series(c), dd=array([[1, 2], [3, 4]]))
    {
        "b": {
            "d": {
                "x": 3
            },
            "y": 5,
            "df": {0: {0: 1, 1: 3}, 1: {0: 2, 1: 4}},
            "ell": "..."
        },
        "z": 9,
        "c": [1 2 3],
        "d": {0: 1, 1: 2, 2: 3},
        "dd": [[1 2] [3 4]]
    }
    """

    width = None

    def default(self, obj):
        if obj is not None:
            from ldict.frozenlazydict import FrozenLazyDict
            from ldict.lazyval import LazyVal

            if obj is Ellipsis:
                return "..."
            if isinstance(obj, FrozenLazyDict):
                return self.data
            if isinstance(obj, LazyVal):
                return str(obj)
            # if isinstance(obj, FunctionType):
            #     return str(obj)
            if not isinstance(obj, (list, set, str, int, float, bytearray, bool)):
                try:
                    from pandas.core.frame import DataFrame, Series

                    if isinstance(obj, (DataFrame, Series)):
                        # «str()» is to avoid nested identation
                        return truncate("«" + str(obj.to_dict()) + "»", self.width)
                    from numpy import ndarray

                    if isinstance(obj, ndarray):
                        return truncate("«" + str(obj).replace("\n", "") + "»", self.width)
                except ImportError:  # pragma: no cover
                    print("Pandas or numpy may be missing.")
                if hasattr(obj, "asdict"):
                    return obj.asdict
                elif hasattr(obj, "aslist"):
                    return obj.aslist
                else:
                    return str(obj)
        return JSONEncoder.default(self, obj)


# class CustomJSONDecoder(JSONDecoder):
#     def __init__(self, *args, **kwargs):
#         JSONDecoder.__init__(self, object_hook=self.object_hook, *args, **kwargs)
#
#     def object_hook(self, obj):
#         if obj is not None:
#             if isinstance(obj, str) and len(obj) == digits:
#                 return
#         return obj


def truncate(txt, width):
    return txt[:width] + "..." if width and len(txt) > width else txt



#  Copyright (c) 2021. Davi Pereira dos Santos
#  This file is part of the ldict project.
#  Please respect the license - more about this in the section (*) below.
#
#  ldict is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ldict is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with ldict.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is illegal and unethical regarding the effort and
#  time spent here.
#
import json
import operator
from functools import reduce, cached_property
from random import Random
from typing import Dict, TypeVar, Union, Callable

from ldict.core.appearance import decolorize

from ldict.core.base import AbstractLazyDict
from ldict.core.rshift import handle_dict, lazify
from ldict.customjson import CustomJSONEncoder
from ldict.exception import WrongKeyType, ReadOnlyLdict
from ldict.lazyval import LazyVal
from ldict.parameter.functionspace import FunctionSpace
from ldict.parameter.let import AbstractLet

VT = TypeVar("VT")


class FrozenLazyDict(AbstractLazyDict):
    """Immutable lazy dict for serializable (picklable) pairs str->value

    Usage:

    >>> from ldict.frozenlazydict import FrozenLazyDict as ldict
    >>> ldict()
    {}
    >>> d = ldict(x=5, y=3)
    >>> d
    {
        "x": 5,
        "y": 3
    }
    >>> d["y"]
    3
    >>> ldict(x=123123, y=88)
    {
        "x": 123123,
        "y": 88
    }
    >>> ldict(y=88, x=123123)
    {
        "y": 88,
        "x": 123123
    }
    >>> d = ldict(x=123123, y=88)
    >>> e = d >> (lambda x: {"z": x**2}) >> (lambda x,y: {"w": x/y})
    >>> e
    {
        "x": 123123,
        "y": 88,
        "z": "→(x)",
        "w": "→(x y)"
    }
    >>> a = d >> (lambda x: {"z": x**2}) >> (lambda x, y: {"w": x/y})
    >>> b = d >> (lambda x, y: {"w": x/y}) >> (lambda x: {"z": x**2})
    >>> dic = d.asdict  # Converting to dict
    >>> dic
    {'x': 123123, 'y': 88}
    >>> d2 = ldict(dic)  # Reconstructing from a dict
    >>> print(d2)
    {
        "x": 123123,
        "y": 88
    }
    >>> from ldict import empty
    >>> d = empty >> {"x": "more content"}
    >>> d
    {
        "x": "more content"
    }
    """

    # noinspection PyMissingConstructor
    def __init__(self, /, _dictionary=None, rnd=None, _returned=None, **kwargs):
        self.rnd = rnd
        self.returned = _returned
        self.data = _dictionary or {}
        self.data.update(kwargs)

    def __getitem__(self, item):
        if not isinstance(item, str):
            raise WrongKeyType(f"Key must be string, not {type(item)}.", item)
        if item not in self.data:
            raise KeyError(item)
        if isinstance(content := self.data[item], LazyVal):
            self.data[item] = content()
        return self.data[item]

    def __setitem__(self, key: str, value):
        del self[key]  # Reuse 'del' exception.

    def __delitem__(self, key):
        raise ReadOnlyLdict(f"Cannot change a frozen dict.", key)

    def __getattr__(self, item):
        if item in self:
            return self[item]
        return self.__getattribute__(item)

    def __repr__(self):
        txt = json.dumps(self.data, indent=4, ensure_ascii=False, cls=CustomJSONEncoder)
        return txt.replace('"«', "").replace('»"', "")

    def __str__(self):
        return decolorize(repr(self))

    @cached_property
    def asdict(self):
        """
        >>> from ldict.frozenlazydict import FrozenLazyDict as ldict
        >>> d = ldict(x=3, y=5)
        >>> ldict(x=7, y=8, d=d).asdict
        {'x': 7, 'y': 8, 'd': {'x': 3, 'y': 5}}
        """
        dic = {}
        for field in self:
            v = self[field]
            dic[field] = v.asdict if isinstance(v, AbstractLazyDict) else v
        return dic

    def clone(self, data=None, rnd=None, _returned=None):
        """Same lazy content with (optional) new data or rnd object."""
        return FrozenLazyDict(self.data if data is None else data, rnd=rnd or self.rnd, _returned=_returned)

    def __rrshift__(self, left: Union[Random, Dict, Callable, FunctionSpace]):
        """
        >>> {"x":5} >> FrozenLazyDict()
        {
            "x": 5
        }
        >>> (lambda x:x*2) >> FrozenLazyDict()
        «λ × {}»
        >>> Random() >> FrozenLazyDict()
        {}
        """
        if isinstance(left, Random):
            return self.clone(rnd=left)
        if isinstance(left, Dict) and not isinstance(left, AbstractLazyDict):
            return FrozenLazyDict(left) >> self
        if callable(left):
            return FunctionSpace(left, self)
        return NotImplemented

    def __rshift__(self, other: Union[Dict, AbstractLazyDict, Callable, AbstractLet, FunctionSpace, Random]):
        from ldict import Empty

        if isinstance(other, Empty):
            return self
        if isinstance(other, Random):
            return self.clone(rnd=other)
        if isinstance(other, FrozenLazyDict):
            return self.clone(handle_dict(self.data, other, other.rnd), other.rnd)
        if isinstance(other, Dict):
            return self.clone(handle_dict(self.data, other, self.rnd))
        if isinstance(other, FunctionSpace):
            return reduce(operator.rshift, (self,) + other.functions)
        if callable(other) or isinstance(other, AbstractLet):
            lazies = lazify(self.data, output_field="extract", f=other, rnd=self.rnd, is_multi_output=True)
            if lazies is None:
                return self
            data = self.data.copy()
            data.update(lazies)
            return self.clone(data, _returned=list(lazies.keys()))
        return NotImplemented

    def __eq__(self, other):
        if self.keys() != other.keys():
            return False
        self.evaluate()
        if isinstance(other, AbstractLazyDict):
            other.evaluate()
            return self.data == other.data
        if isinstance(other, Dict):
            return self.data == other
        raise TypeError(f"Cannot compare {type(self)} and {type(other)}")  # pragma: no cover



#  Copyright (c) 2021. Davi Pereira dos Santos
#  This file is part of the ldict project.
#  Please respect the license - more about this in the section (*) below.
#
#  ldict is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ldict is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with ldict.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is illegal and unethical regarding the effort and
#  time spent here.
#


def check_access(id, readonly, key):
    if readonly:  # pragma: no cover
        raise ReadOnlyLdict(f"Cannot change a readonly ldict ({id}).", key)
    if not isinstance(key, str):
        raise WrongKeyType(f"Key must be string, not {type(key)}.", key)


class OverwriteException(Exception):
    pass


class DependenceException(Exception):
    pass


class NoInputException(Exception):
    pass


class FunctionETypeException(Exception):
    pass


class FromØException(Exception):
    pass


class MissingField(Exception):
    pass


class NoReturnException(Exception):
    pass


class BadOutput(Exception):
    pass


class WrongKeyType(Exception):
    pass


class WrongValueType(Exception):
    pass


class ConflictingParameter(Exception):
    pass


class InconsistentLange(Exception):
    pass


class EmptyNextToGlobalCache(Exception):
    pass


class MultipleIdsForFunction(Exception):
    pass


class MissingIds(Exception):
    pass


class WrongId(Exception):
    pass


class MissingLibraryDependence(Exception):
    pass


class DTypeCannotBeObject(Exception):
    pass


class UnderscoreInField(Exception):
    pass


class MultipleDicts(Exception):
    pass


class UndefinedSeed(Exception):
    pass


class ReadOnlyLdict(Exception):
    pass



#  Copyright (c) 2021. Davi Pereira dos Santos
#  This file is part of the ldict project.
#  Please respect the license - more about this in the section (*) below.
#
#  ldict is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ldict is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with ldict.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is illegal and unethical regarding the effort and
#  time spent here.
from functools import cached_property
from json import dumps
from random import Random
from typing import Callable, Dict, Union

from ldict.core.base import AbstractLazyDict
from ldict.customjson import CustomJSONEncoder
from ldict.parameter.functionspace import FunctionSpace


class AbstractLet:
    def __init__(self, f, dict_type, **kwargs):
        self.f = f
        self.dict_type = dict_type
        self.config = kwargs

    @cached_property
    def asdict(self):
        return self.config

    def __rrshift__(self, left: Union[Dict, Callable, "AbstractLet"]):
        """
        >>> from ldict import empty
        >>> {"x": 5, "y": 7} >> (Random(0) >> (empty >> (lambda x:{"y": 0})))
        {
            "x": 5,
            "y": "→(x)"
        }
        """
        if isinstance(left, dict) and not isinstance(left, AbstractLazyDict):
            return self.dict_type(left) >> self
        if callable(left):
            # noinspection PyArgumentList
            return FunctionSpace(self.__class__(left), self)
        if isinstance(left, (Random, list)):
            return FunctionSpace(self.dict_type(), left, self)
        return NotImplemented  # pragma: no cover

    def __rshift__(self, other: Union[Dict, Callable]):
        if callable(other) or isinstance(other, (list, Random, AbstractLet, Dict)):
            return FunctionSpace(self, other)
        return NotImplemented  # pragma: no cover

    __mul__ = __rshift__

    def __repr__(self):
        return "λ" + str(self.config)



from .core.ldict_ import Ldict as ldict
from .empty import Empty
from .parameter.functionspace import FunctionSpace
from .parameter.let import lLet as let

empty = Empty()
"""The empty object is used to induce a ldict from a dict"""



#  Copyright (c) 2021. Davi Pereira dos Santos
#  This file is part of the ldict project.
#  Please respect the license - more about this in the section (*) below.
#
#  ldict is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ldict is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with ldict.  If not, see <http://www.gnu.org/licenses/>.
#
#  (*) Removing authorship by any means, e.g. by distribution of derived
#  works or verbatim, obfuscated, compiled or rewritten versions of any
#  part of this work is illegal and unethical regarding the effort and
#  time spent here.
import re
from inspect import signature
from io import StringIO
from pprint import pprint

from uncompyle6.main import decompile
from uncompyle6.semantics.parser_error import ParserError

from ldict.exception import NoInputException, NoReturnException, BadOutput, MultipleDicts


def extract_input(f):
    """
    >>> f = lambda x, y, z=5: None
    >>> extract_input(f)
    ({'x': None, 'y': None}, {'z': 5}, ['translated_input'])
    >>> f.metadata = {"input": {"fields": ["a", "b"], "parameters": {"c": 7}}}
    >>> extract_input(f)
    ({'a': None, 'b': None}, {'c': 7}, {})
    """
    if hasattr(f, "metadata") and "input" in f.metadata:
        fields = {k: None for k in f.metadata["input"]["fields"]} if "fields" in f.metadata["input"] else {}
        hasparams = "parameters" in f.metadata["input"] and f.metadata["input"]["parameters"] is not ...
        parameters = f.metadata["input"]["parameters"] if hasparams else {}
        hasoptional = "optional" in f.metadata["input"] and f.metadata["input"]["optional"] is not ...
        optional = f.metadata["input"]["optional"] if hasoptional else {}
        return fields, parameters, optional
    pars = dict(signature(f).parameters)
    input, parameters, optional = {}, {}, ["translated_input"]
    if "kwargs" in pars:
        del pars["kwargs"]
    for k, v in pars.items():
        if v.default is v.empty:
            input[k] = None
        else:
            val = v.default
            if isinstance(val, str) and "=None" in val:
                val = val.split("=")[0]
                optional.append(val)
            parameters[k] = val
    if not input and not parameters:
        raise NoInputException(f"Missing function input parameters.")
    return input, parameters, optional


def extract_body(f):
    """
    Return a readable code

    Doesn't work well with some functions containing dict comprehensions: raises CodeExtractionException.

    >>> def f(x, y, implicit=["a", "b", "c"]):
    ...     return x*y, x+y, x/y
    >>> extract_body(f)
    ['return (x * y, x + y, x / y)']
    """
    if hasattr(f, "metadata") and "code" in f.metadata and f.metadata["code"] is not ...:
        return f.metadata["code"]
    out = StringIO()
    try:
        decompile(bytecode_version=(3, 8, 10), co=f.__code__, out=out)
    except ParserError:
        raise CodeExtractionException("Could not extract function code.")
    code = [line for line in out.getvalue().split("\n") if not line.startswith("#")]
    return code or None


def extract_returnstr(code) -> str:
    """
    >>> def f(x, y, implicit=["a", "b", "c"]):
    ...     return x*y, x+y, x/y
    >>> extract_returnstr("".join(extract_body(f)))
    '(x * y, x + y, x / y)'
    """
    if "return" not in code:
        raise NoReturnException(f"Missing return statement:", code)
    strs = re.findall("(?<=return )(.+)", code)
    if len(strs) != 1:  # pragma: no cover
        raise BadOutput("Cannot detect return expression.", strs)
    return strs[0]


def extract_dictstr(returnstr: str) -> str:
    """
    >>> def f(x, y, implicit=["a", "b", "c"]):
    ...     return {
    ...         "z": x*y,
    ...         "w": x+y,
    ...         implicit: x/y
    ...     }
    >>> extract_dictstr(extract_returnstr("".join(extract_body(f))))
    "{'z': x * y,  'w': x + y,  implicit: x / y}"
    """
    dict_strs = re.findall("(?={)(.+?)(?<=})", returnstr)
    if len(dict_strs) == 0:
        raise BadOutput(
            "Cannot detect output fields, or missing dict (with proper pairs 'identifier'->result) as a return value.",
            dict_strs,
        )
    if len(dict_strs) > 1:
        raise MultipleDicts("Cannot detect output fields, multiple dicts as a return value.", dict_strs)
    return dict_strs[0]


def extract_output(f, body, deps, ismulti_output, dynamic):
    """Extract output fields.

    https://stackoverflow.com/a/68753149/9681577

    >>> extract_output(lambda:None, "return {'z': x*y, 'w': x+y, implicitfield: y**2, '_history': ..., '_code': ..., '_metafield2': 'some text'}", {"implicitfield": "k"}, True, dynamic=[])
    (['z', 'w', 'k'], ['_metafield2'], ['_history', '_code'])
    """

    memo = [""]

    def lazy_dictstr():
        memo[0] = extract_returnstr("".join(body))
        if ismulti_output:
            memo[0] = extract_dictstr(memo[0])
        return memo[0]

    metadata_output = f.metadata["output"] if hasattr(f, "metadata") and "output" in f.metadata else {}
    if "fields" in metadata_output:
        explicit = metadata_output["fields"].copy()
    else:
        explicit = re.findall(r"[\"']([a-zA-Z]+[_a-zA-Z0-9]*)[\"']:", lazy_dictstr())
    if "auto" in metadata_output:
        meta_ellipsed = metadata_output["auto"]
    else:
        meta_ellipsed = re.findall(r"[\"'](_[_a-zA-Z]+[_a-zA-Z0-9]*)[\"']:[ ]*?\.\.\.[,}]", lazy_dictstr())
        meta_ellipsed.extend(re.findall(r"[\"'](_[_a-zA-Z]+[_a-zA-Z0-9]*)[\"']:[ ]*?Ellipsis[,}]", lazy_dictstr()))
    if "meta" in metadata_output:
        meta = metadata_output["meta"]
    else:
        meta = re.findall(r"[\"'](_[_a-zA-Z]+[_a-zA-Z0-9]*)[\"']:", lazy_dictstr())
        meta = [item for item in meta if item not in meta_ellipsed]
    if not dynamic:
        # REMINDER: The variable brings the field name. E.g.: Xout="X"
        dynamic = re.findall(r"[ {]([_a-zA-Z]+[_a-zA-Z0-9]*):", lazy_dictstr())
        # multidynamic = re.findall(r"[ {]kwargs\[([_a-zA-Z]+[_a-zA-Z0-9]*)\[[^]]+?]]:", lazy_dictstr())
        # dynamic.extend(multidynamic)

    for field in dynamic:
        # if "_" in field:  # pragma: no cover
        #     raise UnderscoreInField("Field names cannot contain underscores:", field, dictstr)
        if field not in deps:  # pragma: no cover
            raise Exception("Missing parameter providing implicit field", field, deps)
        if isinstance(deps[field], list):
            explicit.extend(deps[field])
        else:
            explicit.append(deps[field])
    if not explicit:  # pragma: no cover
        pprint(lazy_dictstr())
        raise BadOutput("Could not find output fields that are valid identifiers (or kwargs[...]):")
    return explicit, meta, meta_ellipsed


def extract_dynamic_input(bodystr):
    """The variable brings the field name, so we can get the field content from kwargs.

    >>> extract_dynamic_input("some code kwargs[x].asd * kwargs[y] == 5 kwargs[dyn[23rgf]] ")
    ['dyn', 'x', 'y']
    """
    single = set(re.findall(r"kwargs\[([a-zA-Z]+[a-zA-Z0-9_]*)][^:]", bodystr))
    multi = re.findall(r"kwargs\[([a-zA-Z]+[a-zA-Z0-9_]*)\[[^]]+?]][^:]", bodystr)
    single.update(multi)
    return sorted(list(single))


class CodeExtractionException(Exception):
    pass
